/*  NAME:
        Collier_ImageGradient_GaussianFilter.cpp
 
    DESCRIPTION:
        GaussianFilter class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"


#include "collier_lib.h"
#include "collier_GradientGaussian.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------

// d/dx matrix
int nDX[] = {0,  0, 0,
             1, -2, 1,
             0,  0, 0};

// d/dy matrix
int nDY[] = {0,  1, 0,
             0, -2, 0,
             0,  1, 0};

//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

GradientGaussianFilter::GradientGaussianFilter(double theScale) : ImageOperator()
{
	float **fTempMatrix;
	float *fTempArray;

	//
	// set up gaussian filter
	//
	dSigma_ = theScale;
	wWidth_ = ceil((short)dSigma_)*5;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0.0;
	
	isDoubleOperator_ = TRUE;
	GaussRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = GaussRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i];
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i];
 	}
	
	//
	// set output raster ptr to NULL for safety
	//
	OutRasterF_ = (RasterImage_gray_F *) NULL;	
	 
	
} // end GradientGaussianFilter::GradientGaussianFilter(double theScale)


GradientGaussianFilter::GradientGaussianFilter(short theWidth) : ImageOperator()
{
	float **fTempMatrix;
	float *fTempArray;
	
	//
	// set up gaussian filter
	//
	wWidth_ = theWidth;
	dSigma_ = ((float)wWidth_)/5.0;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0;

	isDoubleOperator_ = TRUE;
	GaussRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = GaussRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i];
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i];
 	}
	
	//
	// set output raster ptr to NULL for safety
	//
	OutRasterF_ = (RasterImage_gray_F *) NULL;	
	

} // end GradientGaussianFilter::GradientGaussianFilter(short theWidth)



GradientGaussianFilter::~GradientGaussianFilter()
{
	// clean up our temp output spaces if they were allocated
	if (DxRasterF_ != NULL)
	{
		delete DxRasterF_;
	}
	if (DyRasterF_ != NULL)
	{
		delete DyRasterF_;
	}
	if (GaussRasterF_ != NULL)
	{
		delete GaussRasterF_;
	}
	if (OutRasterF_ != NULL)
	{
		delete OutRasterF_;
	}
	if (TempRasterF_ != NULL)
	{
		delete TempRasterF_;
	}


} // end GradientGaussianFilter~()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage GradientGaussianFilter::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getBoundRect()));

} // end GradientGaussianFilter.applyTo(RasterImage *imgIn)




RasterImage GradientGaussianFilter::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (OutRasterF_ != NULL)
	{
		delete []OutRasterF_;
	}
	OutRasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(OutRasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output raster in GaussianFilter::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, OutRasterF_);

	return(OutRasterF_);
	
} // end GradientGaussianFilter.applyTo(RasterImage *imgIn, ImageRect *theRect)




void GradientGaussianFilter::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	applyInto(imgIn, imgIn->getBoundRect(), imgOut);

} // end GradientGaussianFilter.applyInto(RasterImage *imgIn, RasterImage *imgOut)




void GradientGaussianFilter::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	// get a temp raster
	if (TempRasterF_ != NULL)
	{
		delete []TempRasterF_;
	}
	TempRasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(TempRasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating temp raster in ImageGradientGaussianFilter::applyInto()");


	// call generic 2D filter routine to do all the heavy lifting
	
	// apply the gaussian filter
	Apply_2D_Filter_F((RasterImage_gray_F*)imgIn, 
					  (RasterImage_gray_F*)GaussRasterF_,
					  fNormFactor_,
					  (ImageRect*) theRect,
					  (RasterImage_gray_F*)TempRasterF_);
					  
	// apply the D/DX filter
	Apply_2D_Filter_F((RasterImage_gray_F*)TempRasterF_, 
					  (RasterImage_gray_F*)DxRasterF_,
					  1.0F, // d/dx is already normalized
					  (ImageRect*) theRect,
					  (RasterImage_gray_F*)imgOut);
	
	// copy the results back into the temp for d/dy				  
	imgOut->copyinto(TempRasterF_);
					  
	// apply the D/DY filter
	Apply_2D_Filter_F((RasterImage_gray_F*)TempRasterF_, 
					  (RasterImage_gray_F*)DyRasterF_,
					  1.0F, // d/dy is already normalized
					  (ImageRect*) theRect,
					  (RasterImage_gray_F*)imgOut);


} // end GradientGaussianFilter.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
