/*  NAME:
        Collier_ImageGradient_GaussianFilter.cpp
 
    DESCRIPTION:
        GaussianFilter class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_GradientGaussianFilter.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------

// d/dx matrix
static int nDX[] = {0,  0, 0,
                   -1, 0, 1,
                    0,  0, 0};

// d/dy matrix
static int nDY[] = {0, -1, 0,
                    0, 0, 0,
                    0,  1, 0};

//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

GradientGaussianFilter::GradientGaussianFilter(double theScale) : ImageOperator()
{
	float **fTempMatrix;
	float *fTempArray;

	//
	// set up gaussian filter
	//
	dSigma_ = theScale;
	wWidth_ = ceil((short)dSigma_)*5;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0.0;
	
	isDoubleOperator_ = TRUE;
	GaussRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = GaussRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i]/2;
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i]/2;
 	}
	
	//
	// set internal ptrs to NULL for safety
	//
	OutRasterF_ = (RasterImage_gray_F *) NULL;	
	TempRasterF_ = (RasterImage_gray_F *) NULL;
	X_RasterF_ = (RasterImage_gray_F *) NULL;
	Y_RasterF_ = (RasterImage_gray_F *) NULL;
	
} // end GradientGaussianFilter::GradientGaussianFilter(double theScale)


GradientGaussianFilter::GradientGaussianFilter(short theWidth) : ImageOperator()
{
	float **fTempMatrix;
	float *fTempArray;
	
	//
	// set up gaussian filter
	//
	wWidth_ = theWidth;
	dSigma_ = ((float)wWidth_)/5.0;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0;

	isDoubleOperator_ = TRUE;
	GaussRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = GaussRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i]/2;
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i]/2;
 	}
	
	//
	// set internal ptrs to NULL for safety
	//
	OutRasterF_ = (RasterImage_gray_F *) NULL;	
	TempRasterF_ = (RasterImage_gray_F *) NULL;
	X_RasterF_ = (RasterImage_gray_F *) NULL;
	Y_RasterF_ = (RasterImage_gray_F *) NULL;

} // end GradientGaussianFilter::GradientGaussianFilter(short theWidth)



GradientGaussianFilter::~GradientGaussianFilter()
{
	// clean up our temp output spaces if they were allocated
	if (DxRasterF_ 		!= NULL)	delete DxRasterF_;
	if (DyRasterF_ 		!= NULL)	delete DyRasterF_;
	if (GaussRasterF_ 	!= NULL)	delete GaussRasterF_;
	if (TempRasterF_ 	!= NULL) 	delete TempRasterF_;
	if (X_RasterF_ 		!= NULL) 	delete X_RasterF_;
	if (Y_RasterF_ 		!= NULL) 	delete Y_RasterF_;
	if (OutRasterF_ 	!= NULL) 	delete OutRasterF_;


} // end GradientGaussianFilter~()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage *GradientGaussianFilter::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end GradientGaussianFilter.applyTo(RasterImage *imgIn)




RasterImage *GradientGaussianFilter::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (OutRasterF_ != NULL)
	{
		delete []OutRasterF_;
	}
	OutRasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(OutRasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output raster in GaussianFilter::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, OutRasterF_);

	return(OutRasterF_);
	
} // end GradientGaussianFilter.applyTo(RasterImage *imgIn, ImageRect *theRect)




void GradientGaussianFilter::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	applyInto(imgIn, imgIn->getValidRect(), imgOut);

} // end GradientGaussianFilter.applyInto(RasterImage *imgIn, RasterImage *imgOut)




void GradientGaussianFilter::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	float **fAtemp, **fBout;
	ImageMath ImageFilter;

	
	// get temp rasters
	if (TempRasterF_ != NULL) delete TempRasterF_;
	TempRasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(TempRasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating temp raster in ImageGradientGaussianFilter::applyInto()");

	if (X_RasterF_ != NULL) delete X_RasterF_;
	X_RasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(X_RasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating X raster in ImageGradientGaussianFilter::applyInto()");

	if (Y_RasterF_ != NULL) delete Y_RasterF_;
	Y_RasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(Y_RasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating Y raster in ImageGradientGaussianFilter::applyInto()");

	// use theRect to limit the filter on the input
	// save old valid rect, restore at end
	ImageRect	*pRect 	= new ImageRect(imgIn->getValidRect());
	imgIn->setValidRect(theRect);

	// apply the D/DX filter to the input image
	ImageFilter.convolve(imgIn,DxRasterF_,X_RasterF_);
	
	// apply the D/DY filter to the input image
	ImageFilter.convolve(imgIn,DyRasterF_,Y_RasterF_);

	// apply the gaussian filter to D/DX
	ImageFilter.convolve(X_RasterF_,GaussRasterF_,TempRasterF_);

	// apply the gaussian filter to D/DY
	ImageFilter.convolve(Y_RasterF_,GaussRasterF_,imgOut);
					  
	// restore imgIn valid rect and delete our temp space
	imgIn->setValidRect(pRect);
	if (pRect!=NULL) delete pRect;

	int nIo = (imgOut->getValidRect())->getTop();
	int nIn = (imgOut->getValidRect())->getBottom();
	int nJo = (imgOut->getValidRect())->getLeft();
	int nJn = (imgOut->getValidRect())->getRight();

	// add the G*D/DX and G*D/DY to get output image
	// this is sorta like the divergence of the vector field
	// at least it gives us something to compare
	fAtemp = TempRasterF_->getGrayF2D();
	fBout = ((RasterImage_gray_F*)imgOut)->getGrayF2D();

	for (int i=nIo; i<=nIn; i++)
	{
		for (int j=nJo; j<=nJn; j++)
		{
			fBout[i][j] += fAtemp[i][j];
		}
	}
	
	// set the valid rect of the output vector
	// already done in convolve above


} // end GradientGaussianFilter.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
