/*  NAME:
        Collier_HoughTransform_circle.cpp
 
    DESCRIPTION:
        Hough Transform circle class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_HoughTransform_circle.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------


//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

HoughTransform_circle::HoughTransform_circle(int m, int n[], int nMin, int nMax)
			       : ImageHoughTransform(m,  n,  nMin,  nMax)
{ 
	pCircleSearch_ = new HoughCircleSearchInternal_Type[nNumSizes_];
	
	for (int i=0; i<nNumSizes_; i++)
	{
		// set the ID to index + 1
		pCircleSearch_[i].Params.nSearchID = i+1;
		// set the number of accumulator rows
		pCircleSearch_[i].Params.nRows = pnAccumSizes_[0];
		// set the number of accumulator columns
		pCircleSearch_[i].Params.nCols = pnAccumSizes_[1];
		// set the pixel search scale, default=1 for now
		// (can maybe guess the user's intent from accumulator dimensions but not yet)
		pCircleSearch_[i].Params.nPixelScale = 1;
		// set the search radius scaled by the pixel search scale
		pCircleSearch_[i].Params.fRadius = (float)((nMin + i*fDeltaDim_)*pCircleSearch_[i].Params.nPixelScale);

		// set a threshold for detection
		// since pi*R*R entries, assume that if we add above R, we found one
		// no other way to get this info here
		pCircleSearch_[i].Params.fThresPercent = .99; // default set to 99% of max accumulator value		
		pCircleSearch_[i].Params.fThresAbsolute = -1.0; // don't use absolute threshold	
		// copy ptr to accumulator raster created in HoughTransform constructor
		pCircleSearch_[i].pAccumRasterGF = pAccumRastersGF_[i];

		//now generate the X,Y's for the circle
		// create zero centered points then offset when applied
		float fTheta = 0.0f,
			  fDeltaTheta = atan(1/pCircleSearch_[i].Params.fRadius);
	    int nNumPts = (int)(1+360.0/fDeltaTheta);
		int nX, nY, 
		    nLastX = (int)(0 - pCircleSearch_[i].Params.fRadius), 
		    nLastY = 0;
		pCircleSearch_[i].nPoints = 0;
		pCircleSearch_[i].pnXpts = new int[nNumPts];
		pCircleSearch_[i].pnYpts = new int[nNumPts];
		while (fTheta<360.0f)
		{
			nX = (int)(pCircleSearch_[i].Params.fRadius*cos(fTheta));
			nY = (int)(pCircleSearch_[i].Params.fRadius*sin(fTheta));
			if ( (nLastX!=nX) || (nLastY!=nY) )
			{
				pCircleSearch_[i].pnXpts[pCircleSearch_[i].nPoints] = nX;
				pCircleSearch_[i].pnYpts[pCircleSearch_[i].nPoints] = nY;
				pCircleSearch_[i].nPoints++;
				nLastX = nX;
				nLastY = nY;				
			}
			fTheta += fDeltaTheta;
		} // end while more angles to compute X,Y for
		
	} // end loop over number of search param sets
	
} // end HoughTransform_circle(n,m,nMin,nMax)

HoughTransform_circle::HoughTransform_circle(int nNumSearchs, HoughCircleSearch_Type *pHCSparams)
			         : ImageHoughTransform(nNumSearchs)
{ 
	ImageMath	ImageFilter;

	pCircleSearch_ = new HoughCircleSearchInternal_Type[nNumSizes_];

	for (int i=0; i<nNumSizes_; i++)
	{
		// copy the user's input params
		pCircleSearch_[i].Params = pHCSparams[i];
		// alloc an accumulator image
		pAccumRastersGF_[i] = new RasterImage_gray_F((short)pCircleSearch_[i].Params.nRows,
													 (short)pCircleSearch_[i].Params.nCols);	
		pAccumRastersGF_[i]->setValidRect(0,0,pCircleSearch_[i].Params.nRows,
		                                      pCircleSearch_[i].Params.nRows); // reset ValRect for zero fill
		ImageFilter.fillInto(0.0f,pAccumRastersGF_[i]); // zero for edges
	    pAccumRastersGF_[i]->setFloatConversionMode(kRangeFloat);

		// keep a copy of the accumulator raster here as well
		pCircleSearch_[i].pAccumRasterGF = pAccumRastersGF_[i];

		// may need to scale the radius but need to come up with a convention first
		//pCircleSearch_[i].Params.fRadius *= pCircleSearch_[i].Params.nPixelScale);

		//now generate the X,Y's for the circle
		// create zero centered points then offset when applied
		float fTheta = 0.0f,
			  fDeltaTheta = atan(1/pCircleSearch_[i].Params.fRadius);
	    int nNumPts = (int)(1+360.0/fDeltaTheta);
		int nX, nY, 
		    nLastX = (int)(0 - pCircleSearch_[i].Params.fRadius), 
		    nLastY = 0;
		pCircleSearch_[i].nPoints = 0;
		pCircleSearch_[i].pnXpts = new int[nNumPts];
		pCircleSearch_[i].pnYpts = new int[nNumPts];
		while (fTheta<360.0f)
		{
			nX = (int)(pCircleSearch_[i].Params.fRadius*cos(fTheta));
			nY = (int)(pCircleSearch_[i].Params.fRadius*sin(fTheta));
			if ( (nLastX!=nX) || (nLastY!=nY) )
			{
				pCircleSearch_[i].pnXpts[pCircleSearch_[i].nPoints] = nX;
				pCircleSearch_[i].pnYpts[pCircleSearch_[i].nPoints] = nY;
				pCircleSearch_[i].nPoints++;
				nLastX = nX;
				nLastY = nY;				
			}
			fTheta += fDeltaTheta;
		} // end while more angles to compute X,Y for
		
	} // end loop over number of search param sets

} // end HoughTransform_circle(nParams,pHCSparams)


HoughTransform_circle::~HoughTransform_circle()
{
	if (pCircleSearch_ != NULL) delete pCircleSearch_;
	
} // end ~HoughTransform_circle()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------
void *HoughTransform_circle::applyTo(RasterImage_gray *imgIn)
{
	return(applyTo(imgIn,(bool)false));
	
} // end HoughTransform_circle::applyTo(imgIn)

void *HoughTransform_circle::applyTo(RasterImage_gray *imgIn, bool isBinary)
{
	ImageMath 			ImageFilter;
	ImageStats_Type		ImageStats;

	// fail is image is not binary
	FAIL_CONDITION(!isBinary,
                kInvalidRasterType,
                "Input Raster is NOT binary in HoughTransform_circle::applyTo(ImgIn,isBinary)");


	// get valid rect of input image
	int nXo = (imgIn->getValidRect())->getLeft(),
		nXn = (imgIn->getValidRect())->getRight(),
		nYo = (imgIn->getValidRect())->getTop(),
		nYn = (imgIn->getValidRect())->getBottom(),
		nWidth = nXn-nXo+1,
		nHeight = nYn-nYo+1;
		
	// get ptr to 2D raster	
	float** fRaster = ((RasterImage_gray_F*)imgIn)->getGrayF2D();
					
	// build up the accumulators for all radii to be searched
	for (int nYi = nYo; nYi <= nYn; nYi++)
	{
		for (int nXi = nXo; nXi <= nXn; nXi++)
		{
			if (fRaster[nYi][nXi] > 0.0f)
			{
				for (int i=0; i<nNumSizes_; i++)
				{
					float** fAccumRaster = pCircleSearch_[i].pAccumRasterGF->getGrayF2D();
					
					int nXAo = (pCircleSearch_[i].pAccumRasterGF->getValidRect())->getLeft(),
						nXAn = (pCircleSearch_[i].pAccumRasterGF->getValidRect())->getRight(),
						nYAo = (pCircleSearch_[i].pAccumRasterGF->getValidRect())->getTop(),
						nYAn = (pCircleSearch_[i].pAccumRasterGF->getValidRect())->getBottom();
					// translate the x,y based on search pixel scale and accumulator size
					float fYoffset = ((float)pCircleSearch_[i].Params.nRows*(float)nYi)/(float)nHeight;
					float fXoffset = ((float)pCircleSearch_[i].Params.nCols*(float)nXi)/(float)nWidth;
					for (int j=0; j<pCircleSearch_[i].nPoints; j++)
					{
						unsigned int unX = pCircleSearch_[i].pnXpts[j] + fXoffset;
						unsigned int unY = pCircleSearch_[i].pnYpts[j] + fYoffset;
						if ( (nXAo<=unX) && (unX<=nXAn) && (nYAo<=unY) && (unY<=nYAn) )
						{
			   				fAccumRaster[unY][unX] += 1;
						}
					} // loop over pts to add into the accumulator
				} // loop over all search sets
			} // if this x,y is on an edge
		} // loop over input X's
	} // loop over input Y's
	
	// delete any old detections
	int nNumResults = 0;
	HCresults_[nNumResults].nSearchID = -1; // in case we don't find any


	// look through the accumulators for detections
	for (int i=0; i<nNumSizes_; i++)
	{
		fRaster = pAccumRastersGF_[i]->getGrayF2D();
		ImageFilter.getStats(pAccumRastersGF_[i], &ImageStats);	

		for (int nYi=0; nYi<pCircleSearch_[i].Params.nRows; nYi++)
		{
			for (int nXi=0; nXi<pCircleSearch_[i].Params.nCols; nXi++)
			{
				if (fRaster[nYi][nXi] > (ImageStats.fMax*pCircleSearch_[i].Params.fThresPercent))
				{
					HCresults_[nNumResults].nSearchID = pCircleSearch_[i].Params.nSearchID;
					HCresults_[nNumResults].nRows = pCircleSearch_[i].Params.nRows;
					HCresults_[nNumResults].nCols = pCircleSearch_[i].Params.nCols;
					HCresults_[nNumResults].fRadius = pCircleSearch_[i].Params.fRadius;
					HCresults_[nNumResults].nXc = nXi;
					HCresults_[nNumResults].nYc = nYi;
					HCresults_[nNumResults].fSum = fRaster[nYi][nXi];
					HCresults_[nNumResults].pRasterGF = pAccumRastersGF_[i];
					nNumResults++;
					HCresults_[nNumResults].nSearchID = -1;
					if (nNumResults==100)
					{
						HCresults_[nNumResults].nSearchID = -2; // tag it as out of space
						return((void *)HCresults_);
					}
				} // if we found a detection
			} // end loop over cols
		} // end loop over rows
	} // loop over all search sets
	
	
	return((void *)HCresults_);
	
} // end HoughTransform_circle::applyTo(imgIn,isBinary)
