/*  NAME:
        Collier_ImageFFT.cpp
 
    DESCRIPTION:
        ImageFFT class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"


#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_ImageFFT.h"
#include "OOURA_alloc.h"
#include "OOURA_fft4f2d.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------


//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

ImageFFT::ImageFFT(int nSize) : ImageOperator()
{
	int n, nw, nc;
	
	nDirection_ = 0; // make the user set it
	
	// make sure size is a power of 2
	FAIL_CONDITION(!IsPowerOf2(nSize),
                kRasterInvalidDimensions,
                "Only 2^n accepted for size ImageFFT::ImageFFT(int nSize)");

    nSize_ = nSize;	  // declared size for this FFT
//
// Forward direction weights
//
	// allocate and make weights
    pnFwdIp_ = alloc_1d_int(2 + (int) sqrt(nSize_ + 0.5));
    alloc_error_check(pnFwdIp_);
    pnFwdIp_[0] = 0;

    pfFwdWeights_ = alloc_1d_float(nSize_*3/2);
    alloc_error_check(pfFwdWeights_);
    
    n = nSize_ << 1;
    if (n < nSize_) { // from OOURA which allowed differing dimensions on X and Y
        n = nSize_;
    }
    
    nw = pnFwdIp_[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        makewt_f(nw, pnFwdIp_, pfFwdWeights_);
    }
    
    nc = pnFwdIp_[1];
    if (n > (nc << 2)) {
        nc = n >> 2;
        makect_f(nc, pnFwdIp_, pfFwdWeights_ + nw);
    }

//
// Inverse direction weights
//
	// allocate and make weights
    pnInvIp_ = alloc_1d_int(2 + (int) sqrt(nSize_ + 0.5));
    alloc_error_check(pnInvIp_);
    pnInvIp_[0] = 0;

    pfInvWeights_ = alloc_1d_float(nSize_*3/2);
    alloc_error_check(pfInvWeights_);
    
    n = nSize_ << 1;
    if (n < nSize_) { // from OOURA which allowed differing dimensions on X and Y
        n = nSize_;
    }
    
    nw = pnInvIp_[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        makewt_f(nw, pnInvIp_, pfInvWeights_);
    }
    
    nc = pnInvIp_[1];
    if (n > (nc << 2)) {
        nc = n >> 2;
        makect_f(nc, pnInvIp_, pfInvWeights_ + nw);
    }
	
	
} // end ImageFFT::ImageFFT(int nSize)



ImageFFT::~ImageFFT()
{
	// clean up our temp output spaces if they were allocated
	if (pnFwdIp_ != NULL)
	{
		delete pnFwdIp_;
	}
	if (pnInvIp_ != NULL)
	{
		delete pnInvIp_;
	}

	if (pfFwdWeights_ != NULL)
	{
		delete pfFwdWeights_;
	}
	if (pfInvWeights_ != NULL)
	{
		delete pfInvWeights_;
	}


} // end ImageFFT::~ImageFFT()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage *ImageFFT::applyTo(RasterImage *imgIn)
{
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end ImageFFT.applyTo(RasterImage *imgIn)


RasterImage *ImageFFT::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	ImageMath ImageFilter;

	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input raster accepted in ImageFFT.applyTo()");

	// make sure we have a valid FFT direction
	FAIL_CONDITION(nDirection_==0,
                kRasterImageError,
                "Direction must be + or - in ImageFFT.applyTo()");
	int nDir = (nDirection_ >0)?1:-1;
	
     // make sure entire image is valid
     // otherwise, we would have to copy it into a temp space
     // use other versions for this
	ImageRect *pValidRectIn   = imgIn->getValidRect(),
		      *pBoundRectIn   = theRect;
    int nVXo   = pValidRectIn->getLeft(),
		nVXn   = pValidRectIn->getRight(),
		nVYo   = pValidRectIn->getTop(),
		nVYn   = pValidRectIn->getBottom();
	int nBXo   = pBoundRectIn->getLeft(),
		nBXn   = pBoundRectIn->getRight(),
		nBYo   = pBoundRectIn->getTop(),
		nBYn   = pBoundRectIn->getBottom();
            
    FAIL_CONDITION( (nVXo!=nBXo)||(nVXn!=nBXn)||(nVYo!=nBYo)||(nVYn!=nBYn),
                kRasterImageError,
                "Input image Valid and defined rects do not match in ImageFFT.applyTo()");    

    // make sure image matches constructed size
    FAIL_CONDITION( (pValidRectIn->getHeight()!=nSize_)||(pValidRectIn->getWidth()!=nSize_),
                kRasterImageError,
                "Input image Valid Rect does not match nSize in ImageFFT.applyTo()");

	// for all floats
	 if (imgIn->hasFloatRaster())
	 {
		float** fA = ((RasterImage_gray_F*)imgIn)->getGrayF2D();
	
     	if (nDir>0)
     	{
    		rdft2d_f(nSize_, nSize_, nDir, fA, pnFwdIp_, pfFwdWeights_);
     		ImageFilter.scale(imgIn,sqrt(2.0)/(2.0f*nSize_));
    	}
    	else if (nDir<0)
    	{
    	    rdft2d_f(nSize_, nSize_, nDir, fA, pnInvIp_, pfInvWeights_);
     	ImageFilter.scale(imgIn,2.0/(nSize_*nSize_));
    	}
    	else
    	{
    		// we should never get here since we failed above
    	}
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only float rasters accepted in ImageFFT.applyTo()");
     }
     	
	 return(imgIn);
	
} // end ImageFFT.applyTo(RasterImage *imgIn, ImageRect *theRect)




void ImageFFT::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	ImageMath ImageFilter;

	imgIn->setValidRect(imgIn->getValidRect());

	ImageFilter.copy(imgIn,imgOut); // non converting copy
	
	RasterImage *pRI = applyTo(imgOut, imgOut->getValidRect());

} // end ImageFFT.applyInto(RasterImage *imgIn, RasterImage *imgOut)



void ImageFFT::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	ImageMath ImageFilter;

	// ignore theRect for now
	ImageRect *tempRect = theRect; // nowarning code
	
	imgIn->setValidRect(imgIn->getValidRect());
	ImageFilter.copy(imgIn,imgOut); // non converting copy
	
	RasterImage *pTempRaster = applyTo(imgOut, imgOut->getValidRect());

} // end ImageFFT.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)


void ImageFFT::applyInto(RasterImage *imgIn, int nPaddedSize, RasterImage *imgOut)
{
	ImageMath	ImageFilter;

    // make sure padded size > imgIn size
    FAIL_CONDITION( (imgIn->getHeight()>nPaddedSize)||(imgIn->getWidth()>nPaddedSize),
                kRasterImageError,
                "Input image Valid Rect too large for padded size in ImageFFT.applyTo()");

	RasterImage *pTempRaster = new RasterImage_gray_F(nPaddedSize,nPaddedSize);
	
	if (pTempRaster!=NULL)
	{
		ImageFilter.copy(imgIn,pTempRaster);
		pTempRaster->setValidRect(pTempRaster->getBoundRect());

		RasterImage *pRI = applyTo(pTempRaster);

		pTempRaster->setValidRect(imgOut->getValidRect());
		ImageFilter.copy(pTempRaster,imgOut);
		imgOut->setValidRect(imgOut->getBoundRect());
	
		delete pTempRaster;
	
	}

} // end ImageFFT.applyInto(RasterImage *imgIn, int nPaddedSize, RasterImage *imgOut)


void ImageFFT::setForward(void)
{
	nDirection_ = 1;
} // end ImageFFT.setForward()


void ImageFFT::setInverse(void)
{
	nDirection_ = -1;
} // end ImageFFT.SetInverse()