/*  NAME:
        Collier_ImageZeroCrossing.cpp
 
    DESCRIPTION:
        Zero Crossing Edge Detector class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_ImageZeroCrossing.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------


//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

ImageZeroCrossing::ImageZeroCrossing(ZeroCrossingMode eMode) : ImageOperator()
{
	eCrossMode_ = eMode;
	 
} // end ImageZeroCrossing()


ImageZeroCrossing::~ImageZeroCrossing()
{
} // end ~ImageZeroCrossing()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage *ImageZeroCrossing::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end ImageZeroCrossing.applyTo(RasterImage *imgIn)




RasterImage *ImageZeroCrossing::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (fOutRaster_ != NULL)
	{
		delete []fOutRaster_;
	}
	fOutRaster_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(fOutRaster_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output raster in LaPlacianFilter::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, fOutRaster_);

	return(fOutRaster_);
	
} // end ImageZeroCrossing.applyTo(RasterImage *imgIn, ImageRect *theRect)




void ImageZeroCrossing::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	applyInto(imgIn, imgIn->getValidRect(), imgOut);

} // end ImageZeroCrossing.applyInto(RasterImage *imgIn, RasterImage *imgOut)




void ImageZeroCrossing::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	int i;
	int j;


    // make sure output image is at least as large as input
    FAIL_CONDITION(imgOut->getHeight() < imgIn->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in ImageZeroCrossEdges::applyInto(In,Rect,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgIn->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in ImageZeroCrossEdges::applyInto(In,Rect,Out)");

	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input raster accepted in ImageZeroCrossEdges::applyInto(In,Rect,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output raster accepted in ImageZeroCrossEdges::applyInto(In,Rect,Out)");

	int nXo = theRect->getLeft(), 
		nYo = theRect->getTop(),
		nXn = theRect->getRight(), 
		nYn = theRect->getBottom();

	switch (eCrossMode_)
	{

	case kFull:

		// for all floats
		if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
		{
			
			float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
			float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
			
			//*************************************************************
			// do upper left cell of valid region
			//*************************************************************
			i = nYo;
			j = nXo;
			fOut[i][j] = 0.0f;
			if ( (fIn[i][j]>0) && // check S, SE, and E neightbors if I'm above 0
			     ( (fIn[i+1][j]<0) || (fIn[i+1][j+1]<0) || (fIn[i][j+1]<0) ) ) fOut[i][j] = 1.0f;
			if ( (fIn[i][j]<0) && // check S, SE, and E neightbors if I'm below 0
			     ( (fIn[i+1][j]>0) || (fIn[i+1][j+1]>0) || (fIn[i][j+1]>0) ) ) fOut[i][j] = 1.0f;
	
			//*************************************************************
			// do left column of valid region except first and last cell
			//*************************************************************
			for (i = nYo+1; i<nYn; i++)
			{
				j = nXo; 
				fOut[i][j] = 0.0f;
				if ( (fIn[i][j]>0) && // check S, SE, E, NE, N neighbors only if I'm above 0
					 ( (fIn[i+1][j]<0) || (fIn[i+1][j+1]<0) || (fIn[i][j+1]<0) || (fIn[i-1][j+1]<0) || (fIn[i-1][j]<0) ) ) fOut[i][j] = 1.0f;
				if ( (fIn[i][j]<0) && // check S, SE, E, NE, N neighbors only if I'm below 0
					 ( (fIn[i+1][j]>0) || (fIn[i+1][j+1]>0) || (fIn[i][j+1]>0) || (fIn[i-1][j+1]>0) || (fIn[i-1][j]>0) ) )fOut[i][j] = 1.0f;
			}

			//*************************************************************
			// do lower left cell of valid region
			//*************************************************************
			i = nYn;
			j = nXo;
			fOut[i][j] = 0.0f;
			if ( (fIn[i][j]>0) && // check E, NE, and N neightbors if I'm above 0
			     ( (fIn[i][j+1]<0) || (fIn[i-1][j+1]<0) || (fIn[i-1][j]<0) ) ) fOut[i][j] = 1.0f;
			if ( (fIn[i][j]<0) && // check E, NE, and N neightbors if I'm below 0
			     ( (fIn[i][j+1]>0) || (fIn[i-1][j+1]>0) || (fIn[i-1][j]>0) ) ) fOut[i][j] = 1.0f;
		

			//*************************************************************
			// do lower row of valid region except first and last cell
			//*************************************************************
			for (j = nXo+1; j<nXn; j++)
			{
				i = nYn;
				fOut[i][j] = 0.0f;
				if ( (fIn[i][j]>0) && // check W, NW, N, NE, E neighbors only if I'm above 0
					 ( (fIn[i][j-1]<0) || (fIn[i-1][j-1]<0) || (fIn[i-1][j]<0) || (fIn[i-1][j+1]<0) || (fIn[i][j+1]<0) ) ) fOut[i][j] = 1.0f;
				if ( (fIn[i][j]<0) && // check W, NW, N, NE, E neighbors only if I'm below 0
					 ( (fIn[i][j-1]>0) || (fIn[i-1][j-1]>0) || (fIn[i-1][j]>0) || (fIn[i-1][j+1]>0) || (fIn[i][j+1]>0) ) ) fOut[i][j] = 1.0f;
			}
		
			//*************************************************************
			// do lower right cell of valid region
			//*************************************************************
			i = nYn;
			j = nXn;
			fOut[i][j] = 0.0f;
			if ( (fIn[i][j]>0) && // check W, NW, and N neightbors if I'm above 0
			     ( (fIn[i][j-1]<0) || (fIn[i-1][j-1]<0) || (fIn[i-1][j]<0) ) ) fOut[i][j] = 1.0f;
			if ( (fIn[i][j]<0) && // check W, NW, and N neightbors if I'm below 0
			     ( (fIn[i][j-1]>0) || (fIn[i-1][j-1]>0) || (fIn[i-1][j]>0) ) ) fOut[i][j] = 1.0f;

			//*************************************************************
			// do right column of valid region except first and last cell
			//*************************************************************
			for (i = nYo+1; i<nYn; i++)
			{
				j = nXn; 
				fOut[i][j] = 0.0f;
				if ( (fIn[i][j]>0) && // check S, SW, W, NW, N neighbors only if I'm above 0
					 ( (fIn[i+1][j]<0) || (fIn[i+1][j-1]<0) || (fIn[i][j-1]<0) || (fIn[i-1][j-1]<0) || (fIn[i-1][j]<0) ) ) fOut[i][j] = 1.0f;
				if ( (fIn[i][j]<0) && // check S, SW, W, NW, N neighbors only if I'm below 0
					 ( (fIn[i+1][j]>0) || (fIn[i+1][j-1]>0) || (fIn[i][j-1]>0) || (fIn[i-1][j-1]>0) || (fIn[i-1][j]>0) ) ) fOut[i][j] = 1.0f;
			}

			//*************************************************************
			// do upper right cell of valid region
			//*************************************************************
			i = nYo;
			j = nXn;
			fOut[i][j] = 0.0f;
			if ( (fIn[i][j]>0) && // check W, SW, and S neightbors if I'm above 0
			     ( (fIn[i][j-1]<0) || (fIn[i+1][j-1]<0) || (fIn[i+1][j]<0) ) ) fOut[i][j] = 1.0f;
			if ( (fIn[i][j]<0) && // check W, SW, and S neightbors if I'm below 0
			     ( (fIn[i][j-1]>0) || (fIn[i+1][j-1]>0) || (fIn[i+1][j]>0) ) ) fOut[i][j] = 1.0f;

			//*************************************************************
			// do top row of valid region except first and last cell
			//*************************************************************
			for (j = nXo+1; j<nXn; j++)
			{
				i = nYo;
				fOut[i][j] = 0.0f;
				
				if ( (fIn[i][j]>0) && // check W, SW, S, SE, E neighbors only if I'm above 0
				     ( (fIn[i][j-1]<0) || (fIn[i+1][j-1]<0) || (fIn[i+1][j]<0) || (fIn[i+1][j+1]<0) || (fIn[i][j+1]<0) ) ) fOut[i][j] = 1.0f;
				if ( (fIn[i][j]<0) && // check W, SW, S, SE, E neighbors only if I'm below 0
					 ( (fIn[i][j-1]>0) || (fIn[i+1][j-1]>0) || (fIn[i+1][j]>0) || (fIn[i+1][j+1]>0) || (fIn[i][j+1]>0) ) ) fOut[i][j] = 1.0f;
			}
		
			//*************************************************************
			// do [1,1] to [n-2,n-2] of valid region
			//*************************************************************
			for (i = nYo+1; i < nYn; i++) 
			{
				for (j = nXo+1; j < nXn; j++) 
				{
					fOut[i][j] = 0.0f;
					if ( (fIn[i][j]>0) && // check W, SW, S, SE, E, NE, N, NW neightbors if I'm above 0
					     ( (fIn[i][j-1]<0) || (fIn[i+1][j-1]<0) || (fIn[i+1][j]<0) ||(fIn[i+1][j+1]<0) ||
					       (fIn[i][j+1]<0) || (fIn[i-1][j+1]<0) || (fIn[i-1][j]<0) ||(fIn[i-1][j-1]<0) )
					     ) fOut[i][j] = 1.0f;
					if ( (fIn[i][j]<0) && // check W, SW, S, SE, E, NE, N, NW neightbors if I'm below 0
					     ( (fIn[i][j-1]>0) || (fIn[i+1][j-1]>0) || (fIn[i+1][j]>0) ||(fIn[i+1][j+1]>0) ||
					       (fIn[i][j+1]>0) || (fIn[i-1][j+1]>0) || (fIn[i-1][j]>0) ||(fIn[i-1][j-1]>0) )
					     ) fOut[i][j] = 1.0f;
				}
			}
		 }
		 else
		 {
		 	// mismatched types, FAIL for now
			FAIL_CONDITION(TRUE,
	                kRasterImageError,
	                "Illegal raster in ImageZeroCrossEdges::applyInto(In,Rect,Out)");
	     }
		break;
		
	case kNeighborsOnly:
		
		// for all floats
		if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
		{
 
 			float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
			float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();

			//*************************************************************
			// do [1,1] to [n-2,n-2] of valid region
			//*************************************************************
			for (i = nYo+1; i < nYn; i++) 
			{
				for (j = nXo+1; j < nXn; j++) 
				{
					fOut[i][j] = 0.0f;
					// check if W, and E have opposite signs
					if ( fIn[i][j-1]*fIn[i][j+1]<0) fOut[i][j] = 1.0f;
					// check if N, and S have opposite signs
					if ( fIn[i-1][j]*fIn[i+1][j]<0) fOut[i][j] = 1.0f;
					// check if NW, and SE have opposite signs
					if ( fIn[i-1][j-1]*fIn[i+1][j+1]<0) fOut[i][j] = 1.0f;
					// check if SW, and NE have opposite signs
					if ( fIn[i+1][j-1]*fIn[i-1][j+1]<0) fOut[i][j] = 1.0f;
				}
			}
		 }
		 else
		 {
		 	// mismatched types, FAIL for now
			FAIL_CONDITION(TRUE,
	                kRasterImageError,
	                "Illegal raster in ImageZeroCrossEdges::applyInto(In,Rect,Out)");
	     }
		break;
		
		default:
		break;
		
	} // end switch over eCrossMode_
		
	// set valid rect for output image
	imgOut->setValidRect(theRect);
	
} // end ImageZeroCrossing.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
